
#include <Keypad.h>
#include <Wire.h>
#include <LiquidCrystal_I2C.h>
#include "RTClib.h"

/* ----------------------------- BUZZER SETUP CORRECT PASSWORD ---------------*/
#define NOTE_B0  31
#define NOTE_C1  33
#define NOTE_CS1 35
#define NOTE_D1  37
#define NOTE_DS1 39
#define NOTE_E1  41
#define NOTE_F1  44
#define NOTE_FS1 46
#define NOTE_G1  49
#define NOTE_GS1 52
#define NOTE_A1  55
#define NOTE_AS1 58
#define NOTE_B1  62
#define NOTE_C2  65
#define NOTE_CS2 69
#define NOTE_D2  73
#define NOTE_DS2 78
#define NOTE_E2  82
#define NOTE_F2  87
#define NOTE_FS2 93
#define NOTE_G2  98
#define NOTE_GS2 104
#define NOTE_A2  110
#define NOTE_AS2 117
#define NOTE_B2  123
#define NOTE_C3  131
#define NOTE_CS3 139
#define NOTE_D3  147
#define NOTE_DS3 156
#define NOTE_E3  165
#define NOTE_F3  175
#define NOTE_FS3 185
#define NOTE_G3  196
#define NOTE_GS3 208
#define NOTE_A3  220
#define NOTE_AS3 233
#define NOTE_B3  247
#define NOTE_C4  262
#define NOTE_CS4 277
#define NOTE_D4  294
#define NOTE_DS4 311
#define NOTE_E4  330
#define NOTE_F4  349
#define NOTE_FS4 370
#define NOTE_G4  392
#define NOTE_GS4 415
#define NOTE_A4  440
#define NOTE_AS4 466
#define NOTE_B4  494
#define NOTE_C5  523
#define NOTE_CS5 554
#define NOTE_D5  587
#define NOTE_DS5 622
#define NOTE_E5  659
#define NOTE_F5  698
#define NOTE_FS5 740
#define NOTE_G5  784
#define NOTE_GS5 831
#define NOTE_A5  880
#define NOTE_AS5 932
#define NOTE_B5  988
#define NOTE_C6  1047
#define NOTE_CS6 1109
#define NOTE_D6  1175
#define NOTE_DS6 1245
#define NOTE_E6  1319
#define NOTE_F6  1397
#define NOTE_FS6 1480
#define NOTE_G6  1568
#define NOTE_GS6 1661
#define NOTE_A6  1760
#define NOTE_AS6 1865
#define NOTE_B6  1976
#define NOTE_C7  2093
#define NOTE_CS7 2217
#define NOTE_D7  2349
#define NOTE_DS7 2489
#define NOTE_E7  2637
#define NOTE_F7  2794
#define NOTE_FS7 2960
#define NOTE_G7  3136
#define NOTE_GS7 3322
#define NOTE_A7  3520
#define NOTE_AS7 3729
#define NOTE_B7  3951
#define NOTE_C8  4186
#define NOTE_CS8 4435
#define NOTE_D8  4699
#define NOTE_DS8 4978
#define REST      0


// change this to make the song slower or faster
int tempo = 114;

/* Set the buzzer pin.*/
int buzzer = 10;

// notes of the moledy followed by the duration.
// a 4 means a quarter note, 8 an eighteenth , 16 sixteenth, so on
// !!negative numbers are used to represent dotted notes,
// so -4 means a dotted quarter note, that is, a quarter plus an eighteenth!!
int melody[] = {
  NOTE_G4,8, NOTE_C4,8, NOTE_E4,16, NOTE_F4,16, NOTE_G4,8, NOTE_C4,8, NOTE_E4,16, NOTE_F4,16,
  NOTE_G4,8, NOTE_C4,8, NOTE_E4,16, NOTE_F4,16, NOTE_G4,8, NOTE_C4,8, NOTE_E4,16, NOTE_F4,16,
  NOTE_G4,-4, NOTE_C4,-4,//5

  NOTE_DS4,16, NOTE_F4,16, NOTE_G4,4, NOTE_C4,4, NOTE_DS4,16, NOTE_F4,16, //6
  NOTE_D4,-1, //7 and 8
  NOTE_F4,-4, NOTE_AS3,-4,
  NOTE_DS4,16, NOTE_D4,16, NOTE_F4,4, NOTE_AS3,-4,
  NOTE_DS4,16, NOTE_D4,16, NOTE_C4,-4 //11 and 12  
};

// sizeof gives the number of bytes, each int value is composed of two bytes (16 bits)
// there are two values per note (pitch and duration), so for each note there are four bytes
int notes = sizeof(melody) / sizeof(melody[0]) / 2;

// this calculates the duration of a whole note in ms
int wholenote = (60000 * 4) / tempo;

int divider = 0, noteDuration = 0;

/* RTC module setup.*/
RTC_DS3231 rtc;

void RTC_module_handler() {
  if (!rtc.begin()) {
    Serial.println("Couldn't find RTC");
    Serial.flush();
    abort();
  }

  if (rtc.lostPower()) {
    Serial.println("RTC lost power, let's set the time!");
    rtc.adjust(DateTime(F(__DATE__), F(__TIME__)));
    // This line sets the RTC with an explicit date & time, for example to set
    // January 21, 2014 at 3am you would call:
    // rtc.adjust(DateTime(2014, 1, 21, 3, 0, 0));
  }
}

void correct_password_handler() {
  // iterate over the notes of the melody.
  // Remember, the array is twice the number of notes (notes + durations)
  for (int thisNote = 0; thisNote < notes * 2; thisNote = thisNote + 2) {

    // calculates the duration of each note
    divider = melody[thisNote + 1];
    if (divider > 0) {
      // regular note, just proceed
      noteDuration = (wholenote) / divider;
    } else if (divider < 0) {
      // dotted notes are represented with negative durations!!
      noteDuration = (wholenote) / abs(divider);
      noteDuration *= 1.5; // increases the duration in half for dotted notes
    }

    // we only play the note for 90% of the duration, leaving 10% as a pause
    tone(buzzer, melody[thisNote], noteDuration * 0.9);

    // Wait for the specief duration before playing the next note.
    delay(noteDuration);

    // stop the waveform generation before the next note.
    noTone(buzzer);
  }
}

/* ---------------------------------------------------------------------------*/

/* Compute the keypad matrix.*/
const byte ROWS_NUMB = 4;
const byte COLS_NUMB = 4;

char keys[ROWS_NUMB][COLS_NUMB] = { {'1', '2', '3', 'A'},
                                    {'4', '5', '6', 'B'},
                                    {'7', '8', '9', 'C'},
                                    {'*', '0', '#', 'D'}
                                  };
/* Set the rows pins and the columns pins for the arduino. The keypad will be connected to
 *  the #9 - #2 of the arduino.
 */
byte row_pins[ROWS_NUMB] = {9, 8, 7, 6};
byte col_pins[COLS_NUMB] = {5, 4, 3, 2};

/* Pins for leds and solenoid lock.*/
int red_led = 11;
int green_led = 12;
int solenoid = A3; 

/* We had 6 keys corresponding to each charcater of the password.*/
char key1, key2, key3, key4, key5, key6, begining_key, reset_key;

/* ----------------------------  The password --------------------------  */
char pass_code[6];


/* Set the D input for the "enter" command and "C" for close the locker*/
char dump_key, final_key;
char enter_key[] = "D";
char close_locker_key;
char off[] = "C";

/* Mark the character wrong.*/
boolean wrong_password = false;
int tries_number = 0;
char buff[7];
int char_wrong = -1;
int number_chars_wrong = 0;

/* Compute the keypad object.*/
Keypad keypad = Keypad(makeKeymap(keys), row_pins, col_pins, ROWS_NUMB, COLS_NUMB);

/* Set the LCD address to 0x27 for a 16 chars and 2 line display.*/
LiquidCrystal_I2C lcd(0x27,16,2);

DateTime now;

void setup() {
  Serial.begin(9600);
  memset(buff, '\0', 6);
  pinMode(red_led, OUTPUT);
  pinMode(green_led, OUTPUT);
  pinMode(buzzer, OUTPUT);
  pinMode(solenoid, OUTPUT);

  RTC_module_handler();
  now = rtc.now();
  char moment[10];
  sprintf(moment, "%02hhu:%02hhu:%02hhu", now.hour(), now.minute(), now.second());

  /* Set the safety deposit box password according to the moment of the day (hour:minutes:seconds).*/
  char password[6];
  set_password(moment, pass_code);
}

/* Function that computes the digits sum for a number.*/
int get_digits_sum(int nr) {
  int sum = 0;
  while (nr > 0) {
    sum += (nr % 10);
    nr /= 10;
  }

  return sum;
}

/* Function that returns power of base.*/
long powint(int base, int exponent) {
  if (exponent == 0) {
    return 1;
  } else {
    return powint(base, exponent - 1) * base;
  }
}

/* 
 *  Function that sets a password for locker according to the moment of the day
 *  returned by rtc module.
 */
void set_password(char *moment, char password[6]) {
  Serial.println(moment);

  char pass_0, pass_1, pass_2, pass_3, pass_4, pass_5;

  /* Get first character of the password.*/
  int first_ch = ((moment[0] - '0') + (moment[1] - '0') +
                    (moment[3] - '0') + (moment[4] - '0')) << 1;

  if (first_ch > 9) {
    if (first_ch == 10) {
      pass_0 = 'A';
    } else if (first_ch == 11) {
      pass_0 = 'B';
    } else if (first_ch == 12) {
      pass_0 = 'C';
    } else if (first_ch == 13) {
      pass_0 = 'D';
    } else if (first_ch == 14) {
      pass_0 = '*';
    } else if (first_ch == 15) {
      pass_0 = '#'; 
    } else {
      int digits_sum = get_digits_sum(first_ch);
      while (digits_sum > 9) {
        int sum = get_digits_sum(digits_sum);
        digits_sum = sum;
      }
      pass_0 = digits_sum + '0';
    }
  } else {
    pass_0 = '0' + first_ch;
  }
  
  /* Get second charcater of the password.*/
  int sec = (moment[6] - '0') * 10 + (moment[7] - '0');
  if (sec > 30) {
    pass_1 = '#';
  } else {
    pass_1 = '*';
  }

  /* Get the third charcater of the password.*/
  int day_hour = (moment[0] - '0') * 10 + (moment[1] - '0');
  if (day_hour % 2 == 0) {
    if (day_hour % 3 == 0) {
      pass_2 = '1';
    } else if (day_hour % 5 == 0) {
      pass_2 = '0';
    } else if (day_hour % 7 == 0) {
      pass_2 = '2';
    } else if (day_hour % 13 == 0) {
      pass_2 = '3';
    } else {
      pass_2 = '4';
    }
  } else if (day_hour % 3 == 0) {
    if (day_hour % 5 == 0) {
      pass_2 = '6';
    } else if (day_hour % 7 == 0) {
      pass_2 = '7';
    } else if (day_hour % 13 == 0) {
      pass_2 = '9';
    } else {
      pass_2 = '8';
    } 
  } else if (day_hour % 5 == 0) {
    if (day_hour % 7 == 0) {
      pass_2 = 'A';
    } else if (day_hour % 13 == 0) {
      pass_2 = 'B';
    } else {
      pass_2 = 'C';
    }
  } else if (day_hour % 7 == 0) {
    pass_2 = (day_hour % 13) ? '*' : '#';
  } else if (day_hour % 13 == 0) {
    pass_2 = '5';
  } else {
    pass_2 = 'D';
  }

  /* Get the fourth character of the password.(hour is AM or PM)*/
  if (day_hour < 12) {
    pass_3 = 'A';
  } else {
    pass_3 = 'B';
  }


  /* Get the fifth character of the password. */
  int day_minutes = (moment[3] - '0') * 10 + (moment[4] - '0');
  if (day_minutes > 30) {
    if (day_minutes == 10) {
      pass_4 = 'A';
    } else if (day_minutes == 11) {
      pass_4 = 'B';
    } else if (day_minutes == 12) {
      pass_4 = 'C';
    } else if (day_minutes == 13) {
      pass_4 = 'C';
    } else {
      if (day_minutes <= 9) {
        pass_4 = day_minutes + '0';
      } else {
        pass_4 = (((day_minutes - 5) * 2) % 10) + '0';
      }
    }
  } else {
    pass_4 = ((day_minutes * 4) % 7) + '0';
  }

  /* Get the final character of the password.*/
  long power = powint(2, day_hour) + 5;
  int rest = power % 16;
  switch (rest)
  {
    case 0:
      pass_5 = 'A';
      break;
    case 1:
      pass_5 = 'C';
      break;
    case 2:
      pass_5 = '*';
      break;
    case 3:
      pass_5 = '0';
      break;
    case 4:
      pass_5 = '2';
      break;
    case 5:
      pass_5 = '3';
      break;
    case 6:
      pass_5 = '9';
      break;
    case 7:
      pass_5 = '8';
      break;
    case 8:
      pass_5 = '4';
      break;
    case 9:
      pass_5 = 'B';
      break;
    case 10:
      pass_5 = 'D';
      break;
    case 11:
      pass_5 = '1';
      break;
    case 12:
      pass_5 = '5';
      break;
    case 13:
      pass_5 = '7';
      break;
    case 14:
      pass_5 = '#';
      break;
    case 15:
      pass_5 = '6';
      break;
  }
  /* Compute the entire password.*/
  sprintf(password, "%c%c%c%c%c%c", pass_0, pass_1, pass_2,
                                    pass_3, pass_4, pass_5);
  Serial.println(password);
}

/* Function that initialize the lcd setup and print first messages.*/
void lcd_setup() {
  lcd.begin(16, 2);
  lcd.backlight();
  lcd.clear();

  lcd.setCursor(4, 0);
  lcd.print("Welcome!");

  delay(2000);
  lcd.clear();
  lcd.setCursor(4, 0);
  lcd.print("I am your");
  lcd.setCursor(0, 1);
  lcd.print("safe-deposit box");
  
  delay(3000);

  lcd.clear();
  lcd.setCursor(2, 0);
  lcd.print("These are the");
  lcd.setCursor(5, 1);
  lcd.print("rules:");

  delay(3000);

  lcd.clear();
  lcd.setCursor(4, 0);
  lcd.print("1. Only 6");
  lcd.setCursor(2, 1);
  lcd.print("characters :)");
  
  delay(2000);

  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("2. Introduce the");
  lcd.setCursor(0, 1);
  lcd.print("correct password");
  delay(2000);
  lcd.clear();
  lcd.setCursor(2, 0);
  lcd.print("to unlock the");
  lcd.setCursor(3, 1);
  lcd.print("deposit box.");
  
  delay(3000);
  
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("3. Press D for");
  lcd.setCursor(1,1);
  lcd.print("verifying it ;)");
  
  delay(3000);
}


void loop() {
  lcd_setup();
 
enter_password:
  noTone(buzzer);
  wrong_password = false;

  /* Set the LCD parameters.*/
  lcd.clear();
  lcd.setCursor(5, 0);
  lcd.print("Enter");
  lcd.setCursor(3, 1);
  lcd.print("password:");

first_character:
  number_chars_wrong = 0;
  /* Get the first character from the keypad.*/
  key1 = keypad.waitForKey();
  
  /* First character input --------------------------------------*/
  if (key1 == pass_code[0]) {
    strcpy(buff, "*");
    lcd.clear();
    lcd.setCursor(5, 0);
    lcd.print(buff);
    goto second_character;
  } else {
    wrong_password = true;
    char_wrong = 0;
    number_chars_wrong += 1;
  }

  /* second character input -----------------------------------------*/
second_character:
  key2 = keypad.waitForKey();
  if (key2 == pass_code[1]) {
    strcpy(buff, "**");
    lcd.clear();
    lcd.setCursor(5, 0);
    lcd.print(buff);
    goto third_character;
      
  } else if (key2 == 'C') {
    memset(buff, 0, 7);
    strcpy(buff, "");
    lcd.clear();
    lcd.setCursor(5, 0);
    lcd.print(buff);

    if (char_wrong == 0 and number_chars_wrong == 1) {
      wrong_password = false;
    }
    
    goto first_character;
  } else {
    wrong_password = true;
    strcpy(buff, "**");
    lcd.clear();
    lcd.setCursor(5, 0);
    lcd.print(buff);
    char_wrong = 1;
    number_chars_wrong += 1;
  }

  /* third charcater input -------------------------------------------*/
third_character:
  key3 = keypad.waitForKey();
  if (key3 == pass_code[2]) {
    strcpy(buff, "***");
    lcd.clear();
    lcd.setCursor(5, 0);
    lcd.print(buff);
    goto fourth_character;
  } else if (key3 == 'C') {
    memset(buff, 0, 7);
    strcpy(buff, "*");
    lcd.clear();
    lcd.setCursor(5, 0);
    lcd.print(buff);

    if (char_wrong == 1 and number_chars_wrong == 1) {
      wrong_password = false;
    }
    
    goto second_character; 
  } else {
    wrong_password = true;
    strcpy(buff, "***");
    lcd.clear();
    lcd.setCursor(5, 0);
    lcd.print(buff);
    char_wrong = 2;
    number_chars_wrong += 1;
  }

  /* fourth chracter input -----------------------------------------*/
fourth_character:
  key4 = keypad.waitForKey();
  if (key4 == pass_code[3]) {
    strcpy(buff, "****");
    lcd.clear();
    lcd.setCursor(5, 0);
    lcd.print(buff);
    goto fifth_character;
      
  } else if (key4 == 'C') {
    memset(buff, 0, 7);
    strcpy(buff, "**");
    lcd.clear();
    lcd.setCursor(5, 0);
    lcd.print(buff);

    if (char_wrong == 2 and number_chars_wrong == 1) {
      wrong_password = false;
    }
    goto third_character; 
  } else {
    wrong_password = true;
    strcpy(buff, "****");
    lcd.clear();
    lcd.setCursor(5, 0);
    lcd.print(buff);
    char_wrong = 3;
    number_chars_wrong += 1;
  }

/* fifth character -------------------------------------------------*/
fifth_character:
  key5 = keypad.waitForKey();
  if (key5 == pass_code[4]) {
    strcpy(buff, "*****");
    lcd.clear();
    lcd.setCursor(5, 0);
    lcd.print(buff);
    goto last_character;

  } else if (key5 == 'C') {
    Serial.print("Dcnu");
    memset(buff, 0, 7);
    strcpy(buff, "***");
    lcd.clear();
    lcd.setCursor(5, 0);
    lcd.print(buff);

    if (char_wrong == 3 and number_chars_wrong == 1) {
      wrong_password = false;
    }
    goto fourth_character;
  } else {
    wrong_password = true;
    strcpy(buff, "*****");
    lcd.clear();
    lcd.setCursor(5, 0);
    lcd.print(buff);
    char_wrong = 4;
    number_chars_wrong += 1;
  }

/* sixth character of the password --------------------------------*/
last_character:
  key6 = keypad.waitForKey();
  if (key6 == pass_code[5]) {
      strcpy(buff, "******");
      lcd.clear();
      lcd.setCursor(5, 0);
      lcd.print(buff);
  } else if (key6 == 'C') {
    memset(buff, 0, 7);
    strcpy(buff, "****");
    lcd.clear();
    lcd.setCursor(5, 0);
    lcd.print(buff);

    if (char_wrong == 4 and number_chars_wrong == 1) {
      wrong_password = false;
    }
    goto fifth_character;
  } else {
    wrong_password = true;
    strcpy(buff, "******");
    lcd.clear();
    lcd.setCursor(5, 0);
    lcd.print(buff);
    char_wrong = 5;
    number_chars_wrong += 1;
  }

/* Verify the password: if it is true or false.*/
  if (wrong_password == true) {
    Serial.println("aici");
    wrong_password = false;
    goto error;
  }
 

/* If the password is correct, press the D character for finish it.*/
correct:
  final_key = keypad.waitForKey();

  if (final_key == 'C') {
    memset(buff, 0, 7);
    strcpy(buff, "*****");
    lcd.clear();
    lcd.setCursor(5, 0);
    lcd.print(buff);
    goto last_character;
  } else if (final_key == enter_key[0]) {
    
    /* The green led light is on. */ 
    digitalWrite(green_led, HIGH);
    digitalWrite(solenoid, HIGH);

    lcd.clear();
    lcd.println("Locker activated!");

    correct_password_handler();
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("Press C button");
    lcd.setCursor(3, 1);
    lcd.print("to reset!");

  } else {
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("You have to press");
    lcd.setCursor(0, 1);
    lcd.print("D for finish.");
    goto correct;
  }

/* Close the locker.*/
close_lock:
  close_locker_key = keypad.waitForKey();
  if (close_locker_key == off[0]) {

    /* Locker is deactivated -> green led light id off.*/
    digitalWrite(green_led, LOW);
    digitalWrite(solenoid, LOW);

    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("Deposit closed!");
    goto enter_password;
  } else {
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("Press C to close");
    lcd.setCursor(2, 1);
    lcd.print("the safe box.");
    goto close_lock;
  }

/* Error handler in case we have introduced the wrong password.*/
error:
  final_key = keypad.waitForKey();

  if (final_key == enter_key[0]) {
    tries_number += 1;
    digitalWrite(red_led, HIGH);
    tone(buzzer, 1000);
  
    lcd.clear();
    lcd.setCursor(0,0);
    lcd.print("Wrong password!");
  
    if (tries_number == 4) {
      goto lock_blocked;
    }
  
    delay(2000);
    lcd.clear();
    lcd.setCursor(3, 0);
    lcd.print("Retry after");
    lcd.setCursor(4, 1);
    lcd.print("5 seconds!");
    delay(5000);
    digitalWrite(red_led, LOW);
    goto enter_password;
    
  } else {
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("You have to press");
    lcd.setCursor(0, 1);
    lcd.print("D for finish.");
  }

  
lock_blocked:
  delay(2000);
  lcd.clear();
  lcd.setCursor(0,0);
  lcd.print("You've blocked");
  lcd.setCursor(0, 1);
  lcd.print("the safe deposit!");
  delay(4000);
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print ("Please try again");
  lcd.setCursor(4, 1);
  lcd.print("later!");
  delay(15000);
  digitalWrite(red_led, LOW);
  goto enter_password;
}
